/****************************************************************************
 *
 * MODULE:       r.neighbors.lm
 * AUTHOR(S):    Francisco Alonso-Sarría
 * PURPOSE:      Linear model as a focal operator
 * COPYRIGHT:    
 *
 *               This program is free software under the GNU General Public
 *               License (>=v2). Read the file COPYING that comes with GRASS
 *               for details.
 *
 *****************************************************************************/


#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <grass/gis.h>
#include <grass/raster.h>
#include <grass/glocale.h>
#include <math.h>
#include <stdbool.h>

#define PI 3.141592

int lm(double *xx, double *yy, int n, double *a, double *b, double *r){
    double mx=0, my=0, s2x=0, s2y=0, cv=0;
    int i;
    for (i=0;i<n;i++){ mx = mx + *(xx+i); my = my + *(yy+i);} 
    mx = mx/n; my = my/n;  // Means
    
    for (i=0;i<n;i++) {s2x = s2x + pow(*(xx+i)-mx,2); s2y = s2y + pow(*(yy+i)-my,2);} 
    s2x = s2x/(n-1); s2y = s2y/(n-1);  // Variances
    
    for (i=0;i<n;i++) cv = cv + (*(xx+i)-mx) * (*(yy+i)-my);
    cv = cv/(n-1); // Covariance
    
    *r = cv/(sqrt(s2x)*sqrt(s2y));
    *b = cv/s2x;
    *a = my - (*b) * mx;
    return(0);
}
    

int main(int argc, char *argv[]){

    struct GModule *module;
    struct Option *mapax, *mapay;
    struct Option *mapaa, *mapab, *mapar;
    struct Option *ventana;
    struct Cell_head region;

    // Output maps are DCELL, input map may be any type
    CELL  *datosx_c,*datosy_c;
    FCELL *datosx_f,*datosy_f;
    DCELL *datosx_d,*datosy_d; 
    CELL  *entradax_c, *entraday_c;
    FCELL *entradax_f, *entraday_f;
    DCELL *entradax_d, *entraday_d;
    DCELL *datosa, *datosb, *datosr;
    DCELL *xx, *yy;
    DCELL a,b,r;

    char *name_b, *name_a, *name_r, *name_x, *name_y;   // Nombres de capas

    int tipox, tipoy, row, col, nn, vv, dc, dr, tam, k, k2;  //Varios enteros
    int fda,fdb,fdr,fdx,fdy;  //File identifiers


    // Defining module parameters and options
    G_gisinit(argv[0]);

    module = G_define_module();

    module->description =_("Calculates simple linear regression in windows");
    
    G_add_keyword(_("raster"));
    G_add_keyword(_("focal"));

    mapax=G_define_standard_option(G_OPT_R_INPUT);
    mapax -> key = "indep";
    mapax -> description = "Independent variable map"; 

    mapay=G_define_standard_option(G_OPT_R_INPUT);
    mapay -> key = "dep";
    mapay -> description = "Dependent variable map";

    mapaa=G_define_standard_option(G_OPT_R_OUTPUT);
    mapaa -> key = "intercept";
    mapaa -> description = "Intercept of the regression";

    mapab=G_define_standard_option(G_OPT_R_OUTPUT);
    mapab -> key = "slope";
    mapab -> description = "Slope of the regression";

    mapar = G_define_standard_option(G_OPT_R_OUTPUT);
    mapar -> key = "correlation";
    mapar -> required = NO;
    mapar -> description = "Correlation coefficient";

    ventana=G_define_option();
    ventana -> key = "size";
    ventana -> type = TYPE_INTEGER;
    ventana -> description = "window size";
    ventana -> required = YES;

    if(G_parser(argc, argv)) exit(EXIT_FAILURE);


    name_x = mapax->answer;
    name_y = mapay->answer;
    name_b = mapab->answer;
    name_a = mapaa->answer;
    name_r = mapar->answer;


    // Obtains basic info
    nn=atoi(ventana->answer);
    vv=(nn-1)/2;              
    tam=nn*nn;                 

    G_get_window(&region);

    // Reserves memory for values inside the window
    xx = (DCELL *) G_malloc(tam* sizeof(DCELL));    
    yy = (DCELL *) G_malloc(tam* sizeof(DCELL));     
 
    // Open files and Reserve memory for input layers        
    fdx = Rast_open_old(name_x, ""); tipox=Rast_get_map_type(fdx);     
    fdy = Rast_open_old(name_y, ""); tipoy=Rast_get_map_type(fdy);
    
    switch(tipox){
       case 0:{
          entradax_c=(CELL  *) G_malloc(region.rows*region.cols * sizeof(CELL));  
          datosx_c=(CELL  *) G_malloc(region.cols * sizeof(CELL));  break;
       }
       case 1:{
          entradax_f=(FCELL  *) G_malloc(region.rows*region.cols * sizeof(FCELL));
          datosx_f=(FCELL *) G_malloc(region.cols * sizeof(FCELL)); break;
       }
       case 2:{
          entradax_d=(DCELL  *) G_malloc(region.rows*region.cols * sizeof(DCELL));
          datosx_d=(DCELL *) G_malloc(region.cols * sizeof(DCELL)); break;
       }
    }
    
    switch(tipoy){
       case 0:{
          entraday_c=(CELL  *) G_malloc(region.rows*region.cols * sizeof(CELL));  
          datosy_c=(CELL  *) G_malloc(region.cols * sizeof(CELL));  break;
       }
       case 1:{
          entraday_f=(FCELL  *) G_malloc(region.rows*region.cols * sizeof(FCELL));
          datosy_f=(FCELL *) G_malloc(region.cols * sizeof(FCELL)); break;
       }
       case 2:{
          entraday_d=(DCELL  *) G_malloc(region.rows*region.cols * sizeof(DCELL));
          datosy_d=(DCELL *) G_malloc(region.cols * sizeof(DCELL)); break;
       }
    }

    // Open files and Reserve memory for output layers
    fda=Rast_open_new(name_a,2); 
    datosa = (DCELL *) G_malloc(region.cols * sizeof(DCELL));
    fdb=Rast_open_new(name_b,2); 
    datosb = (DCELL *) G_malloc(region.cols * sizeof(DCELL));
    fdr=Rast_open_new(name_r,2); 
    datosr = (DCELL *) G_malloc(region.cols * sizeof(DCELL));


printf("Read input files: \n");
   // Read input files  
    for (row = 0; row < region.rows; row++) {
        G_percent(row, region.rows - 1, 1);
        switch(tipox){
           case 0: Rast_get_row(fdx,datosx_c,row,tipox);break;
           case 1: Rast_get_row(fdx,datosx_f,row,tipox);break;
           case 2: Rast_get_row(fdx,datosx_d,row,tipox);break;
        }
        switch(tipoy){
           case 0: Rast_get_row(fdy,datosy_c,row,tipoy);break;
           case 1: Rast_get_row(fdy,datosy_f,row,tipoy);break;
           case 2: Rast_get_row(fdy,datosy_d,row,tipoy);break;
        }
        for (col=0;col<region.cols;col++){ 
           switch(tipox){
              case 0: *(entradax_c + row*region.cols + col)= *(datosx_c + col);break;
              case 1: *(entradax_f + row*region.cols + col)= *(datosx_f + col);break;
              case 2: *(entradax_d + row*region.cols + col)= *(datosx_d + col);break;
           }
           switch(tipoy){
              case 0: *(entraday_c + row*region.cols + col)= *(datosy_c + col);break;
              case 1: *(entraday_f + row*region.cols + col)= *(datosy_f + col);break;
              case 2: *(entraday_d + row*region.cols + col)= *(datosy_d + col);break;
           }

        }
    }

printf("Processing:\n");
    for (row = 0; row < region.rows; row++) {
       G_percent(row, region.rows - 1, 1);
       for (col=0;col<region.cols;col++){  
          if (row>=vv && row<(region.rows-vv) && col>=vv && col<(region.cols-vv)){
              k=0;
              for (dr=-vv;dr<=vv;dr++){ 
                for (dc=-vv;dc<=vv;dc++){
                   switch(tipox){
                      case 0: *(xx+k) = (DCELL) *(entradax_c + (row+dr)*region.cols + col+dc); break;
                      case 1: *(xx+k) = (DCELL) *(entradax_f + (row+dr)*region.cols + col+dc); break;
                      case 2: *(xx+k) = (DCELL) *(entradax_d + (row+dr)*region.cols + col+dc); break;
                   }            
                   switch(tipoy){
                      case 0: *(yy+k) = (DCELL) *(entraday_c + (row+dr)*region.cols + col+dc); break;
                      case 1: *(yy+k) = (DCELL) *(entraday_f + (row+dr)*region.cols + col+dc); break;
                      case 2: *(yy+k) = (DCELL) *(entraday_d + (row+dr)*region.cols + col+dc); break;
                   } 
                   k++;
                } 
              }
              k2 = lm(xx, yy, tam, &a, &b, &r);  // Modelo lineal
              *(datosa + col) = a;               // Ponemos los resultados en las 
              *(datosb + col) = b;               // correspondientes fila
              *(datosr + col) = r;
          } else {
              Rast_set_d_null_value (datosa + col, 1);  // Si estamos en los
              Rast_set_d_null_value (datosb + col, 1);  // bordes de la capa
              Rast_set_d_null_value (datosr + col, 1);  // ponemos nulo
          }
       }
       Rast_put_row(fda,datosa,2);
       Rast_put_row(fdb,datosb,2);
       Rast_put_row(fdr,datosr,2);
    }



    // CERRAR Y SALIR
printf("Closing\n");
    Rast_close(fdx);
    Rast_close(fdy);
    Rast_close(fda);
    Rast_close(fdb);
    Rast_close(fdr);

    // Liberar memoria
    switch(tipox){
           case 0:  G_free(datosx_c); G_free(entradax_c); break;
           case 1:  G_free(datosx_f); G_free(entradax_f); break;
           case 2:  G_free(datosx_d); G_free(entradax_d); break;
    }
    switch(tipoy){
           case 0:  G_free(datosy_c); G_free(entraday_c); break;
           case 1:  G_free(datosy_f); G_free(entraday_f); break;
           case 2:  G_free(datosy_d); G_free(entraday_d); break;
    }
    G_free(datosa);G_free(datosb);G_free(datosr);
    return (0);
}
